/*
 * Copyright (C) 2001-2011 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "NmdcHub.h"

#include "ResourceManager.h"
#include "ClientManager.h"
#include "SearchManager.h"
#include "ShareManager.h"
#include "CryptoManager.h"
#include "ConnectionManager.h"

#include "Socket.h"
#include "UserCommand.h"

NmdcHub::NmdcHub(const string& aHubURL) : Client(aHubURL, '|'),
	lastbytesshared(0), validatenicksent(false), bFirstOpList(true),
	PtokaX(false), ShittYnHub(false) { }

NmdcHub::~NmdcHub() throw() {
	clearUsers();
}

void NmdcHub::connect(const OnlineUser& aUser, const string&) {
	checkstate();
	dcdebug("NmdcHub::connect %s\n", aUser.getIdentity().getNick().c_str());
	if(ClientManager::getInstance()->isActive()) {
		connectToMe(aUser);
	} else {
		revConnectToMe(aUser);
	}
}

int64_t NmdcHub::getAvailable() const {
	Lock l(cs);
	int64_t x = 0;
	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
		x+=i->second->getIdentity().getBytesShared();
	}
	return x;
}

OnlineUser& NmdcHub::getUser(const string& aNick) {
	OnlineUser* u = NULL;
	{
		Lock l(cs);

		NickIter i = users.find(aNick);
		if(i != users.end())
			return *i->second;
	}

	UserPtr p;
	if(aNick == getCurrentNick()) {
		p = ClientManager::getInstance()->getMe();
	} else {
		p = ClientManager::getInstance()->getUser(aNick, getHubUrl());
	}

	{
		Lock l(cs);
		u = users.insert(make_pair(aNick, new OnlineUser(p, *this))).first->second;
		u->getIdentity().setNick(aNick);
		if(u->getUser() == getMyIdentity().getUser()) {
			setMyIdentity(u->getIdentity());
		}
	}

	ClientManager::getInstance()->putOnline(u);
	return *u;
}

void NmdcHub::supports(const StringList& feat) {
	string x;
	for(StringList::const_iterator i = feat.begin(); i != feat.end(); ++i) {
		x+= ' ' + *i;
	}
	if(getStealth() == false) {
		send("$Supports" + x + '|');
    } else {
		send("$Supports" + x + " |");
    }
}

OnlineUser* NmdcHub::findUser(const string& aNick) {
	Lock l(cs);
	NickIter i = users.find(aNick);
	return i == users.end() ? NULL : i->second;
}

void NmdcHub::putUser(const string& aNick) {
	OnlineUser* ou = NULL;
	{
		Lock l(cs);
		NickIter i = users.find(aNick);
		if(i == users.end())
			return;
		ou = i->second;
		users.erase(i);
	}
	ClientManager::getInstance()->putOffline(ou);
	delete ou;
}

void NmdcHub::clearUsers() {
	NickMap u2;

	{
		Lock l(cs);
		u2.swap(users);
	}

	for(NickIter i = u2.begin(); i != u2.end(); ++i) {
		ClientManager::getInstance()->putOffline(i->second);
		delete i->second;
	}
	u2.clear();
}

void NmdcHub::updateFromTag(Identity& id) {
    string tag = id.getTag();
    if(tag == Util::emptyString)
        return;

    char *sTag = (char *)tag.c_str();
    sTag[tag.size()-1] = ',';
    char *sPart = sTag+1;
    bool bHaveSpace = false;
    for(size_t i = 1; i < tag.size(); i++) {
        switch(sTag[i]) {
            case ' ':
                if(bHaveSpace == false) {
                    if((sTag+i)-sPart > 0) {
                        string sCL = string(sPart, (sTag+i)-sPart);
                        id.setClientID(sCL);
                    }

                    sPart = sTag+i+1;
                    bHaveSpace = true;
                }
                break;
            case ',':
                if(sPart[1] == ':') {
                    switch(sPart[0]) {
                		case 'V':
                            if((sTag+i)-(sPart+2)) {
                                string sVE = string(sPart+2, (sTag+i)-(sPart+2));
                                id.setVersion(sVE);
                            }

                    		break;
                		case 'M':
                            if((sTag+i)-(sPart+2)) {
                                string sMD = string(sPart+2, (sTag+i)-(sPart+2));
                                id.setMode(sMD);
                            }

                			break;
                		case 'H':
                			if((sTag+i)-(sPart+2) > 4) {
                				int a,b,c;
                				if(sscanf_s(sPart+2, "%d/%d/%d", &a, &b, &c) == 3) {
    			                    id.setHubs(Util::toString(a+b+c));
                				} else {
                                    if((sTag+i)-(sPart+2)) {
                                        string sHT = string(sPart+2, (sTag+i)-(sPart+2));
                                        id.setHubs(sHT);
                                    }
                                }
                			} else {
                                if((sTag+i)-(sPart+2)) {
                                    string sHT = string(sPart+2, (sTag+i)-(sPart+2));
                                    id.setHubs(sHT);
                                }
                			}
                    		break;
                		case 'S':
                            if((sTag+i)-(sPart+2)) {
                                string sSL = string(sPart+2, (sTag+i)-(sPart+2));
                                id.setSlots(sSL);
                            }

                    		break;
                		case 'L':
                        case 'B':
                        case 'U':
                            if((sTag+i)-(sPart+2)) {
                                string sUL = string(sPart+2, (sTag+i)-(sPart+2));
                                id.setUpload(sUL);
                            }

                			break;
                		default:
                       	    break;
                    }
                }

                sPart = sTag+i+1;
                break;
            default:
                continue;
        }
    }
    sTag[tag.size()-1] = '>';
}

void NmdcHub::on(Line, const string& l) throw() {
    Client::on(Line(), l);

    COMMAND_DEBUG(l, DebugManager::HUB_IN, getIpPort());

    char* aLine = (char *)l.c_str();
    size_t iaLineLen = l.size();

	if(l[0] != '$') {
    	// Check if we're being banned...
    	if(state != STATE_LOGGEDIN) {
    		if(strstr(aLine, "banned") != 0) {
    			setAutoReconnect(false);
    		}
    	}

    	bool bNormal = true;
    	OnlineUser* u = findUserInChat(aLine, bNormal);

		fire(ClientListener::Message(), this, *u, TO_UTF8(Util::unescape(bNormal ? string(aLine, iaLineLen) : string(aLine+4, iaLineLen-4))));
        return;
    }

	switch(aLine[1]) {
		case 'Z':
			// $ZOn
			if(iaLineLen == 4 && ((uint16_t *)(aLine+2))[0] == ((uint16_t *)"On")[0]) {
				socket->setMode(BufferedSocket::MODE_ZPIPE);
				return;
			}

			break;
        case 'S':
        	// $Search
        	if(state == STATE_LOGGEDIN && iaLineLen > 8 && memcmp(aLine+2, "earch ", 6) == NULL) {
        	    char *sSearchParts[] = { NULL, NULL, NULL, NULL, NULL, NULL, NULL };
                unsigned short iSearchPartsLen[] = { 0, 0, 0, 0, 0, 0, 0 };

                unsigned char cPart = 0;

                sSearchParts[cPart] = aLine+8; // IP for active search, Hub for passive search...

        	    for(size_t i = 8; i < iaLineLen; i++) {
                    switch(aLine[i]) {
                        case ':':
                            if(cPart == 0) {
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case ' ':
                            if(cPart == 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);
                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                        case '?':
                            if(cPart > 1) {
                                aLine[i] = '\0';
                                iSearchPartsLen[cPart] = (unsigned short)((aLine+i)-sSearchParts[cPart]);

                                // are we on last '?' ?
                                if(cPart == 5) {
                                    sSearchParts[6] = aLine+i+1;
                                    iSearchPartsLen[6] = (unsigned short)(iaLineLen-i-1);
                                    i = iaLineLen;
                                    break;
                                }

                                cPart++;
                                sSearchParts[cPart] = aLine+i+1;
                            }
                            break;
                    }
                }

       	        // check if we have all we need
        	    if(iSearchPartsLen[0] == 0 || iSearchPartsLen[1] == 0 || iSearchPartsLen[2] != 1 || iSearchPartsLen[3] != 1 ||
        	       iSearchPartsLen[4] == 0 || iSearchPartsLen[5] == 0 || iSearchPartsLen[6] == 0)
        	       return;

        		bool bPassive = (iSearchPartsLen[0] == 3 && ((uint32_t *)(sSearchParts[0]))[0] == ((uint32_t *)"Hub:")[0]);

        		// We don't wan't to answer passive searches if we're in passive mode...
        		if(bPassive == true && ClientManager::getInstance()->isActive() == false) {
        			return;
        		}

        		// Filter own searches
        		if(ClientManager::getInstance()->isActive() && bPassive == false) {
                    string sIP = ClientManager::getInstance()->getCachedIp();
                    string sPort = Util::toString(SearchManager::getInstance()->getPort());
                    if(iSearchPartsLen[0] == sIP.size() && strncmp(sSearchParts[0], sIP.c_str(), iSearchPartsLen[0]) == NULL &&
                        iSearchPartsLen[1] == sPort.size() && strcmp(sSearchParts[1], sPort.c_str()) == NULL) {
        				return;
        			}
        		} else {
					string sMyNick = fromUtf8(getCurrentNick());
        			if(iSearchPartsLen[1] == sMyNick.size() && strcmp(sSearchParts[1], sMyNick.c_str()) == 0) {
        				return;
        			}
        		}

        		uint64_t tick = GET_TICK();
        		clearFlooders(tick);

        		string seeker(bPassive == false ? sSearchParts[0] : sSearchParts[0] + 4, 
                    bPassive == false ? iSearchPartsLen[0]+iSearchPartsLen[1]+1 : (iSearchPartsLen[0]+iSearchPartsLen[1])-3);

        		seekers.push_back(make_pair(seeker, tick));

        		// First, check if it's a flooder
        		for(FloodIter fi = flooders.begin(); fi != flooders.end(); ++fi) {
        			if(fi->first == seeker)
        				return;
        		}

        		int count = 0;
        		for(FloodIter fi = seekers.begin(); fi != seekers.end(); ++fi) {
        			if(fi->first == seeker)
        				count++;

        			if(count > 7) {
        				// BFU don't need to know search spammers ;)
        				if(isOp()) {
        					if(bPassive == true) {
        						fire(ClientListener::SearchFlood(), this, TO_UTF8(Util::unescape(string(sSearchParts[1], iSearchPartsLen[1]))));
        					} else {
        						fire(ClientListener::SearchFlood(), this, TO_UTF8(Util::unescape(seeker)) + STRING(NICK_UNKNOWN));
        					}
        				}				
        				flooders.push_back(make_pair(seeker, tick));
        				return;
        			}
        		}

        		int a;
        		if(sSearchParts[2][0] == 'F') {
        			a = SearchManager::SIZE_DONTCARE;
        		} else {
        			if(sSearchParts[3][0] == 'F') {
        				a = SearchManager::SIZE_ATLEAST;
        			} else {
        				a = SearchManager::SIZE_ATMOST;
        			}
        		}

        		int64_t size = _atoi64(sSearchParts[4]);
        		int type = atoi(sSearchParts[5]) - 1;

        		if(bPassive == true) {
        			OnlineUser* u = findUser(TO_UTF8(Util::unescape(string(sSearchParts[1], iSearchPartsLen[1]))));

        			if(u == NULL)
                        return;
                        
                    if(!u->getUser()->isSet(User::PASSIVE)) {
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
        			}
        		}

        		fire(ClientListener::NmdcSearch(), this, seeker, a, size, type, TO_UTF8(Util::unescape(string(sSearchParts[6], iSearchPartsLen[6]))), bPassive);

        		return;
            }

            // $SR
            if(iaLineLen > 4 && ((uint16_t *)(aLine+2))[0] == ((uint16_t *)"R ")[0]) {
    			SearchManager::getInstance()->onSearchResult(TO_UTF8(Util::unescape(l)));
    			return;
            }

            // $Supports
           	if(iaLineLen > 8 && memcmp(aLine+2, "upports", 7) == NULL) {
                char *sSupport = aLine+10;

                for(size_t i = 10; i < iaLineLen; i++) {
                    if(aLine[i] == ' ') {
                        aLine[i] = '\0';
                    } else if(i != iaLineLen-1) {
                        continue;
                    } else {
                        i++;
                    }

                    size_t iDataLen = (aLine+i)-sSupport;

                	switch(sSupport[0]) {
                        case 'U':
                        	if(iDataLen == 11 && memcmp(sSupport+1, "serCommand", 10) == NULL) {
                        		setFlag(SUPPORTS_USERCOMMAND);
                        	} else if(iDataLen == 7 && memcmp(sSupport+1, "serIP2", 6) == NULL) {
                        		setFlag(SUPPORTS_USERIP2);
                            }
                            break;
                        case 'N':
                        	if(iDataLen == 9 && ((uint64_t *)(sSupport+1))[0] == ((uint64_t *)"oGetINFO")[0]) {
                        		setFlag(SUPPORTS_NOGETINFO);
                            }
                            break;
                        case 'Q':
                        	if(iDataLen == 9 && ((uint64_t *)(sSupport+1))[0] == ((uint64_t *)"uickList")[0]) {
                        		if(state == STATE_HELLO) {
                        			state = STATE_MYINFO;
                        			updateCounts(false);
                        			myInfo();
                        			getNickList();
                        			setFlag(SUPPORTS_QUICKLIST);
                        		}
                        	}
                        	break;
                        default:
                            dcdebug("NmdcHub::onLine Unknown supports %s\n", aLine);
                            break;
                    }
                    sSupport = aLine+i+1;
                }

                if(isSet(SUPPORTS_QUICKLIST) == false) {
            		OnlineUser& ou = getUser(getCurrentNick());
            		validateNick(ou.getIdentity().getNick());
            	}
            	return;
            }

            break;
    	case 'M':
        	// $MyINFO $ALL  $ $$$$
        	if(iaLineLen > 20 && memcmp(aLine+2, "yINFO $ALL ", 11) == NULL) {
                char *sMyINFOParts[] = { NULL, NULL, NULL, NULL, NULL, NULL };
                unsigned short iMyINFOPartsLen[] = { 0, 0, 0, 0, 0, 0 };

                unsigned char cPart = 0;

                sMyINFOParts[cPart] = aLine+13; // Nick

                for(size_t i = 13; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
                        aLine[i] = '\0';
                        iMyINFOPartsLen[cPart] = (unsigned short)((aLine+i)-sMyINFOParts[cPart]);

                        // are we on end of myinfo ???
                        if(cPart == 5)
                            break;

                        cPart++;
                        sMyINFOParts[cPart] = aLine+i+1;
                    } else if(cPart == 0 && aLine[i] == ' ') {
                        aLine[i] = '\0';
                        iMyINFOPartsLen[cPart] = (unsigned short)((aLine+i)-sMyINFOParts[cPart]);
                        cPart++;
                        sMyINFOParts[cPart] = aLine+i+1;
                    }
                }

                // check if we have nick...
        	    if(iMyINFOPartsLen[0] == 0)
        	         return;

                string sNick = TO_UTF8(Util::unescape(string(sMyINFOParts[0], iMyINFOPartsLen[0])));
        		OnlineUser& u = getUser(sNick);

        		// nick case sensitivity fix
        		if(strcmp(u.getIdentity().getNick().c_str(), sMyINFOParts[0]) != NULL) {
                    u.getIdentity().set("NI", sNick);
                }

                // clear all tag parts
        		u.getIdentity().setClientID(Util::emptyString);
        		u.getIdentity().setVersion(Util::emptyString);
        		u.getIdentity().setMode(Util::emptyString);
        		u.getIdentity().setHubs(Util::emptyString);
				u.getIdentity().setSlots(Util::emptyString);
        		u.getIdentity().setUpload(Util::emptyString);

                // check if we have description + tag
                if(iMyINFOPartsLen[1] != 0) {
                    if(sMyINFOParts[1][iMyINFOPartsLen[1]-1] == '>') {
                        char *sTag;
        				if((sTag = strrchr(sMyINFOParts[1], '<')) != NULL) {
        					u.getIdentity().setTag(Util::unescape(string(sTag, iMyINFOPartsLen[1]-(sTag-sMyINFOParts[1]))));
        					// Hm, we have something...disassemble it...
							updateFromTag(u.getIdentity());
							if(sTag-sMyINFOParts[1] != 0) {
                                sTag[0] = '\0';
                                u.getIdentity().setDescription(Util::unescape(string(sMyINFOParts[1], sTag-sMyINFOParts[1])));
                            } else {
                                u.getIdentity().setDescription(Util::emptyString);
                            }
        				} else {
							u.getIdentity().setDescription(Util::unescape(string(sMyINFOParts[1], iMyINFOPartsLen[1])));
                            u.getIdentity().setTag(Util::emptyString);
                        }
        			} else {
						u.getIdentity().setDescription(Util::unescape(string(sMyINFOParts[1], iMyINFOPartsLen[1])));
						u.getIdentity().setTag(Util::emptyString);
					}
                } else {
                    u.getIdentity().setDescription(Util::emptyString);
                    u.getIdentity().setTag(Util::emptyString);
                }

                // check if we have 'space' before connection '$'
                if(iMyINFOPartsLen[2] != 0 && sMyINFOParts[2][0] != ' ') {
                    if(sMyINFOParts[2][0] == 'A') {
            			u.getIdentity().getUser()->unsetFlag(User::PASSIVE);
            		} else {
            			u.getIdentity().getUser()->setFlag(User::PASSIVE);
                    }
                }

                // check if we have connection
                if(iMyINFOPartsLen[3] != 0) {
                    char status = sMyINFOParts[3][iMyINFOPartsLen[3]-1];
            		sMyINFOParts[3][iMyINFOPartsLen[3]-1] = '\0';
            		if(iMyINFOPartsLen[3] != 1) {
                        u.getIdentity().setConnection(Util::unescape(string(sMyINFOParts[3], iMyINFOPartsLen[3]-1)));

               			// set connection type
                		switch(sMyINFOParts[3][0]) {
                            case '0':
                                if(sMyINFOParts[3][1] == '.') {
                                    if(((uint32_t *)(sMyINFOParts[3]+2))[0] == ((uint32_t *)"005\0")[0] || 
                                        (sMyINFOParts[3][4] == '\0' && (((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"01")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"02")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"05")[0]))) {
                                        u.getIdentity().setcType(1);
                                        break;
                                    } else if(((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"1\0")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"2\0")[0] || 
                                        ((uint16_t *)(sMyINFOParts[3]+2))[0] == ((uint16_t *)"5\0")[0]) {
                                        u.getIdentity().setcType(5);
                                        break;
                                    }
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '1':
                                if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(5);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if((sMyINFOParts[3][3] == '\0' && ((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"00")[0]) ||
                                    ((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"000\0")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '2':
                                if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                } else if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"8.8Kbps\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '3':
                    			if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"3.6Kbps\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case '5':
                    			if(sMyINFOParts[3][1] == '\0') {
                                    u.getIdentity().setcType(6);
                                    break;
                                } else if(((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"0\0")[0]) {
                                    u.getIdentity().setcType(7);
                                    break;
                                } else if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"6Kbp")[0] && 
                                    ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"s\0")[0]) {
                                    u.getIdentity().setcType(1);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'M':
                                if(sMyINFOParts[3][5] == '\0' && ((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"odem")[0]) {
                    				u.getIdentity().setcType(1);
                    				break;
                                } else if(sMyINFOParts[3][9] == '\0' && ((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"icrowave")[0]) {
                    				u.getIdentity().setcType(3);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'I':
                        		if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"SDN\0")[0]) {
                    				u.getIdentity().setcType(2);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'S':
                        		if(sMyINFOParts[3][9] == '\0' && ((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"atellite")[0]) {
                                    u.getIdentity().setcType(3);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'W':
                                if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"ireless\0")[0]) {
                    				u.getIdentity().setcType(3);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'D':
                        		if(sMyINFOParts[3][3] == '\0' && ((uint16_t *)(sMyINFOParts[3]+1))[0] == ((uint16_t *)"SL")[0]) {
                    				u.getIdentity().setcType(4);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'C':
                        		if(sMyINFOParts[3][5] == '\0' && ((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"able")[0]) {
                    				u.getIdentity().setcType(5);
                    				break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'L':
                                if(sMyINFOParts[3][7] == '\0') {
                            		if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"AN(T")[0] && 
                                        ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"1)")[0]) {
                                        u.getIdentity().setcType(6);
                                        break;
                                    } if(((uint32_t *)(sMyINFOParts[3]+1))[0] == ((uint32_t *)"AN(T")[0] && 
                                        ((uint16_t *)(sMyINFOParts[3]+5))[0] == ((uint16_t *)"3)")[0]) {
                                        u.getIdentity().setcType(7);
                                        break;
                                    }
                                }
                                u.getIdentity().setcType(11);
                                break;
                            case 'N':
                                if(((uint64_t *)(sMyINFOParts[3]+1))[0] == ((uint64_t *)"etlimite")[0] && sMyINFOParts[3][9] == 'r') {
                                    u.getIdentity().setcType(8);
                                    break;
                                }
                                u.getIdentity().setcType(11);
                                break;
                            default:
                                // no known connection type found, set unknown
                                u.getIdentity().setcType(11);
                                break;
                        }
                    } else {
                        u.getIdentity().setConnection(Util::emptyString);
                        u.getIdentity().setcType(11);
                    }

                    // set status based on magic byte
                    if((status & 0x8) == 0x8) {
               			u.getIdentity().setFlag(Identity::FIREBALL);
                    } else {
                        u.getIdentity().unsetFlag(Identity::FIREBALL);
                    }

                    if((status & 0x4) == 0x4) {
               			u.getIdentity().setFlag(Identity::SERVER);
                    } else {
                        u.getIdentity().unsetFlag(Identity::SERVER);
                    }

                    if((status & 0x2) == 0x2) {
              			u.getIdentity().setFlag(Identity::AWAY);
                    } else {
                        u.getIdentity().unsetFlag(Identity::AWAY);
                    }
                } else {
                    u.getIdentity().setConnection(Util::emptyString);
                    u.getIdentity().setcType(11);
                    u.getIdentity().unsetFlag(Identity::FIREBALL);
                    u.getIdentity().unsetFlag(Identity::SERVER);
                    u.getIdentity().unsetFlag(Identity::AWAY);
                }

                // check if we have email
                if(iMyINFOPartsLen[4] != 0) {
                    u.getIdentity().setEmail(Util::unescape(string(sMyINFOParts[4], iMyINFOPartsLen[4])));
                } else {
                    u.getIdentity().setEmail(Util::emptyString);
                }

                // check if we have sharesize
                if(iMyINFOPartsLen[5] != 0) {
                    u.getIdentity().setBytesShared(Util::unescape(string(sMyINFOParts[5], iMyINFOPartsLen[5])));
                } else {
                    u.getIdentity().setBytesShared("0");
                }

        		if(state == STATE_MYINFO) {
					if(u.getUser() == getMyIdentity().getUser()) {
        				state = STATE_LOGGEDIN;
						setMyIdentity(u.getIdentity());
        				updateCounts(false);
        				u.getUser()->setFlag(User::DCPLUSPLUS);
        				if(SETTING(CONNECTION_TYPE) != SettingsManager::CONNECTION_ACTIVE)
        					u.getUser()->setFlag(User::PASSIVE);
        				else
        					u.getUser()->unsetFlag(User::PASSIVE);
        			}
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }

            break;
        case 'Q':
        	// $Quit
    		if(iaLineLen > 6 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"uit ")[0]) {
                string sNick = TO_UTF8(Util::unescape(string(aLine+6, iaLineLen-6)));
        		OnlineUser* u = findUser(sNick);

				if(u == NULL)
                    return;  

        		fire(ClientListener::UserRemoved(), this, *u);
        		putUser(sNick);

        		return;
            }

            break;
        case 'C':
    		// $ConnectToMe
    		if(state == STATE_LOGGEDIN && iaLineLen > 13 && memcmp(aLine+2, "onnectToMe ", 11) == NULL) {
                char *sCTMParts[] = { NULL, NULL, NULL };
                unsigned short iCTMPartsLen[] = { 0, 0, 0 };

                bool bFirst = true;

                sCTMParts[0] = aLine+13; // Nick

                for(size_t i = 13; i < iaLineLen; i++) {
                    if(bFirst == true) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                            iCTMPartsLen[0] = (unsigned short)((aLine+i)-sCTMParts[0]);
                            sCTMParts[1] = aLine+i+1;
                            bFirst = false;
                        }
                    } else if(aLine[i] == ':') {
                        aLine[i] = '\0';
                        iCTMPartsLen[1] = (unsigned short)((aLine+i)-sCTMParts[1]);
                        sCTMParts[2] = aLine+i+1;
                        iCTMPartsLen[2] = (unsigned short)(iaLineLen-i-1);
                        break;
                    }
                }

                // check if we have all we need
                if(iCTMPartsLen[0] == 0 || iCTMPartsLen[1] == 0 || iCTMPartsLen[2] == 0)
                    return;

                 // Check nick... is CTM really for me ? ;o)
				string sMyNick = fromUtf8(getCurrentNick());
        		if(iCTMPartsLen[0] != sMyNick.size() || _stricmp(sCTMParts[0], sMyNick.c_str()) != 0)
        			return;

                string sIP = string(sCTMParts[1], iCTMPartsLen[1]);

                if(inet_addr(sIP.c_str()) != INADDR_NONE) { // Check if IP is really IP and not something else...
                    ConnectionManager::getInstance()->nmdcConnect(sIP, (uint16_t)atoi(sCTMParts[2]), getCurrentNick(), getHubUrl());
                }

        		return;
            }

            break;
        case 'R':
        	// $RevConnectToMe
        	if(state == STATE_LOGGEDIN && iaLineLen > 16 && memcmp(aLine+2, "evConnectToMe ", 14) == NULL) {
				char *sTemp = strchr(aLine+16, ' ');
				if(sTemp == NULL || (sTemp-aLine)-16 == 0 || iaLineLen-(sTemp-aLine)-1 == 0)
					return;

        		// Check nick... is RCTM really for me ? ;o)
				string sMyNick = fromUtf8(getCurrentNick());
        		if(size_t(iaLineLen-(sTemp-aLine)-1) != sMyNick.size() || _stricmp(sTemp+1, sMyNick.c_str()) != 0)
        			return;

        		OnlineUser* u = findUser(TO_UTF8(Util::unescape(string(aLine+16, (sTemp-aLine)-16))));

        		if(u == NULL)
                    return;

        		if(ClientManager::getInstance()->isActive()) {
        			connectToMe(*u);
        			if(u->getUser()->isSet(User::PASSIVE) == false) {
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
                    }
        		} else {
        			// Notify the user that we're passive too...
        			if(u->getUser()->isSet(User::PASSIVE) == false) {
        				if(u->getIdentity().isSet(Identity::RCTM_SND) == false) {
        					revConnectToMe(*u);
        					u->getIdentity().setFlag(Identity::RCTM_SND);
        				}
        				u->getUser()->setFlag(User::PASSIVE);
        				updated(*u);
                    }
        		}

        		return;
            }

            break;
        case 'H':
        	// $HubName
        	if(iaLineLen > 9 && memcmp(aLine+2, "ubName ", 7) == NULL) {
                char *sTmp = strstr(aLine+9, " - ");
        		if(sTmp != NULL && iaLineLen-3-(sTmp-aLine) != 0) {
                    getHubIdentity().setNick(TO_UTF8(Util::unescape(string(aLine+9, sTmp-aLine-9))));
                    getHubIdentity().setDescription(TO_UTF8(Util::unescape(string(sTmp+3, iaLineLen-3-(sTmp-aLine)))));
        		} else {
					getHubIdentity().setNick(TO_UTF8(Util::unescape(string(aLine+9, iaLineLen-9))));
					getHubIdentity().setDescription(Util::emptyString);	
        		}    			
        		fire(ClientListener::HubUpdated(), this);
        		return;
            }

            // $Hello
           	if(isSet(SUPPORTS_QUICKLIST) == false && iaLineLen > 7 && memcmp(aLine+2, "ello ", 5) == NULL) {
        		string nick = TO_UTF8(Util::unescape(string(aLine+7, iaLineLen-7)));
        		OnlineUser& u = getUser(nick);
        		u.getIdentity().set("NI", nick);

        		if(state == STATE_HELLO && u.getUser() == getMyIdentity().getUser()) {
                    state = STATE_LOGGEDIN;
        			updateCounts(false);
            		u.getUser()->setFlag(User::DCPLUSPLUS);
            		if(ClientManager::getInstance()->isActive()) {
            			u.getUser()->unsetFlag(User::PASSIVE);
            		} else {
            			u.getUser()->setFlag(User::PASSIVE);
            		}

        			version();
        			if(ShittYnHub == true || getStealth() == true) {
        				getNickList();
        				myInfo();
                    } else {
        				myInfo();
        				getNickList();
                    }
        		}
				fire(ClientListener::UserUpdated(), this, u);
        		return;
            }

            // $HubIsFull
           	if(iaLineLen == 10 && ((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ubIsFull")[0]) {
				fire(ClientListener::HubFull(), this);
    			return;
        	}

        	break;
        case 'U':
        	// $UserCommand
        	if(iaLineLen > 13 && memcmp(aLine+2, "serCommand ", 11) == NULL) {
                char *sUsrCmdParts[] = { NULL, NULL, NULL, NULL };
                unsigned short iUsrCmdPartsLen[] = { 0, 0, 0, 0 };

                unsigned char cPart = 0;
                sUsrCmdParts[cPart] = aLine+13; // type

                for(size_t i = 13; i < iaLineLen; i++) {
                    switch(cPart) {
                        case 0:
                        case 1:
                            if(aLine[i] == ' ') {
                                aLine[i] = '\0';
                            } else if(i != iaLineLen-1) {
                                continue;
                            } else {
                                i++;
                            }
                            iUsrCmdPartsLen[cPart] = (unsigned short)((aLine+i)-sUsrCmdParts[cPart]);
                            cPart++;
                            sUsrCmdParts[cPart] = aLine+i+1;
                            break;
                        case 2:
                            if(aLine[i] == '$') {
                                aLine[i] = '\0';
                                iUsrCmdPartsLen[2] = (unsigned short)((aLine+i)-sUsrCmdParts[2]);
                                sUsrCmdParts[3] = aLine+i+1;
                                iUsrCmdPartsLen[3] = (unsigned short)(iaLineLen-i-1);
                                i = iaLineLen;
                            }
                            break;
                    }
                }

                // check if we have two needed numbers...
                if(iUsrCmdPartsLen[0] == 0 || iUsrCmdPartsLen[1] == 0)
                    return;

        		int type = atoi(sUsrCmdParts[0]);
        		int ctx = atoi(sUsrCmdParts[1]);;

        		switch(type) {
                    case UserCommand::TYPE_SEPARATOR:
                    case UserCommand::TYPE_CLEAR:
        			    fire(ClientListener::UserCommand(), this, type, ctx, Util::emptyString, Util::emptyString);
        			    break;
            		case UserCommand::TYPE_RAW:
					case UserCommand::TYPE_RAW_ONCE: {
                        // check if we have two needed strings...
                        if(iUsrCmdPartsLen[2] == 0 || iUsrCmdPartsLen[3] == 0)
                            return;

        			    fire(ClientListener::UserCommand(), this, type, ctx, TO_UTF8(Util::unescape(string(sUsrCmdParts[2], iUsrCmdPartsLen[2]))),
                            TO_UTF8(Util::unescape(string(sUsrCmdParts[3], iUsrCmdPartsLen[3]))));
        			    break;
					}
        			default:
                        // ?!?
                        break;
        		}
        		return;
            }

            // $UserIP
           	if(iaLineLen > 8 && memcmp(aLine+2, "serIP ", 6) == NULL) {
				OnlineUser::List v;

        		char *sNick = aLine+8; // first nick
        		char *sIp = NULL;
        		bool bSpace = true;

        		for(size_t i = 8; i < iaLineLen; i++) {
                    if(bSpace == true) {
                        if(aLine[i] == ' ') {
                            aLine[i] = '\0';
                            sIp = aLine+i+1;
                            bSpace = false;
                        }
                    } else {
                        if(aLine[i] == '$') {
                            aLine[i] = '\0';
                        } else if(i != iaLineLen-1) {
                            continue;
                        } else {
                            i++;
                        }

                        string strNick = TO_UTF8(Util::unescape(string(sNick, (sIp-sNick)-1)));
                        OnlineUser& u = getUser(strNick);

                        if(strcmp(u.getIdentity().getNick().c_str(), strNick.c_str()) != NULL) {
                            u.getIdentity().set("NI", strNick);
                        }

                        u.getIdentity().setIp(TO_UTF8(Util::unescape(string(sIp, (aLine+i)-sIp))));
                        v.push_back(&u);

                        sNick = aLine+i+2;
                        i++;
                        bSpace = true;
                    }
                }

				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }

            break;
        case 'L':
        	// $Lock
        	if(state == STATE_LOCK && iaLineLen > 6 && ((uint32_t *)(aLine+2))[0] == ((uint32_t *)"ock ")[0]) {
        		state = STATE_HELLO;
        		string lck;

        		char *sTmp = strstr(aLine+6, " Pk=");
        		if(sTmp != NULL) {
        			lck = string(aLine+6, (sTmp-aLine)-6);
        			if(_stricmp(sTmp+4, "YnHub") == 0) {
        				ShittYnHub = true;
        			} else if(strcmp(sTmp+4, "PtokaX") == 0) {
        				PtokaX = true;
        			}
                } else {
                    lck = l;
                }

        		if(CryptoManager::getInstance()->isExtended(lck)) {
        			StringList feat;
        			feat.push_back("UserCommand"); feat.push_back("NoGetINFO");
            		feat.push_back("NoHello"); feat.push_back("UserIP2");
        			if(getStealth() == false) {
            			feat.push_back("QuickList");
            			feat.push_back("ZPipe");
                    } else {
        				feat.push_back("TTHSearch");
        				feat.push_back("ZPipe");
                    }

        			supports(feat);
					if(PtokaX == false || getStealth() == true) {
                        key(CryptoManager::getInstance()->makeKey(lck));
                    }
        		} else {
        			key(CryptoManager::getInstance()->makeKey(lck));
					OnlineUser& ou = getUser(getCurrentNick());
					validateNick(ou.getIdentity().getNick());
        		}
                return;
    		}

    		// $LogedIn
    		if(iaLineLen > 7 && memcmp(aLine+2, "ogedIn", 6) == NULL) {
    			// nothing to do here...
    			return;
            }

            break;
    	case 'F':
            // $ForceMove
        	if(iaLineLen > 11 && memcmp(aLine+2, "orceMove ", 9) == NULL) {
        		socket->disconnect(false);
        		fire(ClientListener::Redirect(), this, TO_UTF8(Util::unescape(string(aLine+11, iaLineLen-11))));
        		return;
            }

            break;
        case 'V':
        	// $ValidateDenide
        	if(iaLineLen > 14 && memcmp(aLine+2, "alidateDenide", 13) == NULL) {
        		socket->disconnect(false);
				fire(ClientListener::NickTaken(), this);
        		return;
            }

            break;
        case 'N':
        	// $NickList
        	if(iaLineLen > 10 && ((uint64_t *)(aLine+2))[0] == ((uint64_t *)"ickList ")[0]) {
				OnlineUser::List v;
                char *sNick = aLine+10; // first nick

                for(size_t i = 10; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
    					v.push_back(&getUser(TO_UTF8(Util::unescape(string(sNick, (aLine+i)-sNick)))));
                        sNick = aLine+i+2;
                        i++;
                    }
                }

        		if(isSet(SUPPORTS_NOGETINFO) == false && isSet(SUPPORTS_QUICKLIST) == false) {
        			string tmp;
        			// Let's assume 10 characters per nick...
        			tmp.reserve(v.size() * (11 + 10 + getCurrentNick().length())); 
        			string n = ' ' +  fromUtf8(getCurrentNick()) + '|';
					for(OnlineUser::List::const_iterator i = v.begin(); i != v.end(); ++i) {
        				tmp += "$GetINFO ";
        					tmp += fromUtf8((*i)->getIdentity().getNick());
        					tmp += n;
        			}
        			if(!tmp.empty()) {
        				send(tmp);
        			}
        		}
				fire(ClientListener::UsersUpdated(), this, v);
        		return;
            }

            break;
        case 'O':
        	// $OpList
        	if(iaLineLen > 8 && memcmp(aLine+2, "pList ", 6) == NULL) {
				OnlineUser::List v;
                char *sOp = aLine+8; // first OP nick

                for(size_t i = 8; i < iaLineLen; i++) {
                    if(aLine[i] == '$') {
                        string sOpNick = TO_UTF8(Util::unescape(string(sOp, (aLine+i)-sOp)));
                        OnlineUser& ou = getUser(sOpNick);
    					ou.getIdentity().setOp(true);
    					if(isOp() == false && sOpNick == getCurrentNick()) {
    						setMyIdentity(ou.getIdentity());
    					}
    					v.push_back(&ou);
                        sOp = aLine+i+2;
                        i++;
                    }
                }

        		fire(ClientListener::UsersUpdated(), this, v);
        		updateCounts(false);
        		if(bFirstOpList == true) {
        			if(getMyIdentity().isRegistered() == true) {
        				myInfo();
                    }
        			bFirstOpList = false;
        		}
        		return;
            }

            break;
        case 'T':
        	// $To:
        	if(iaLineLen > 5 && memcmp(aLine+2, "o: ", 3) == NULL) {
                char *sToParts[] = { NULL, NULL, NULL, NULL };
                unsigned long iToPartsLen[] = { 0, 0, 0, 0 };

                unsigned char cPart = 0;

                sToParts[cPart] = aLine+5; // Nick

                for(size_t i = 5; i < iaLineLen; i++) {
                    if(aLine[i] == ' ') {
                        aLine[i] = '\0';
                        iToPartsLen[cPart] = (unsigned long)((aLine+i)-sToParts[cPart]);

                        // are we on last space ?
                        if(cPart == 2) {
                            sToParts[3] = aLine+i+2;
                            iToPartsLen[3] = (unsigned long)(iaLineLen-i-2);
                            break;
                        }

                        cPart++;
                        sToParts[cPart] = aLine+i+1;
                    }
                }

                // check if we have all we need
                if(iToPartsLen[0] == 0 || iToPartsLen[1] != 5 || iToPartsLen[2] == 0 || iToPartsLen[3] == 0)
                    return;

                 // Make CrazyGuy happy and check nick... is To really for me ? ;o)
				string sMyNick = fromUtf8(getCurrentNick());
        		if(iToPartsLen[0] != sMyNick.size() || _stricmp(sToParts[0], sMyNick.c_str()) != 0)
        			return;

                bool bNormal = true;
                OnlineUser* by = findUserInChat(sToParts[3], bNormal);

                string sFromNick = TO_UTF8(Util::unescape(string(sToParts[2], iToPartsLen[2])));
				OnlineUser* from = findUser(sFromNick);

				if(from == NULL ) {
					fire(ClientListener::PrivateMessageNick(), this, sFromNick, *by,
                        TO_UTF8(Util::unescape(bNormal ? string(sToParts[3], iToPartsLen[3]) : string(sToParts[3]+4, iToPartsLen[3]-4))));
                } else {	
					fire(ClientListener::PrivateMessage(), this, *from, *by,
                        TO_UTF8(Util::unescape(bNormal ? string(sToParts[3], iToPartsLen[3]) : string(sToParts[3]+4, iToPartsLen[3]-4))));
        		}
        		return;
            }

            break;
        case 'G':
        	// $GetPass
        	if(iaLineLen == 8 && memcmp(aLine+2, "etPass", 6) == NULL) {
				OnlineUser& ou = getUser(getCurrentNick());
				ou.getIdentity().setFlag(Identity::REG);
				setMyIdentity(ou.getIdentity());
        		fire(ClientListener::GetPassword(), this);
        		return;
            }

            break;
        case 'B':
        	// $BadPass
        	if(iaLineLen == 8 && memcmp(aLine+2, "adPass", 6) == NULL) {
        		setPassword(Util::emptyString);
        		return;
            }

            break;
    	default:
		  break;
	}

	dcdebug("NmdcHub::onLine Unknown command %s\n", aLine);
}

string NmdcHub::checkNick(const string& aNick) {
	string tmp = aNick;
	for(size_t i = 0; i < aNick.size(); ++i) {
		switch(tmp[i]) {
			case '|':
			case '$':
			case '<':
			case '>':
				tmp[i] = '_';
			default:
				continue;
		}
	}
	return tmp;
}

void NmdcHub::connectToMe(const OnlineUser& aUser) {
	checkstate(); char buf[256];
	dcdebug("NmdcHub::connectToMe %s\n", aUser.getIdentity().getNick().c_str());
	ConnectionManager::getInstance()->nmdcExpect(aUser.getIdentity().getNick(), getCurrentNick(), getHubUrl());
	int chars = sprintf_s(buf, 256, "$ConnectToMe %s %s:%d|", fromUtf8(aUser.getIdentity().getNick()).c_str(), getLocalIp().c_str(), ConnectionManager::getInstance()->getPort());
	ConnectionManager::iConnToMeCount++;
	if(chars != -1)
		send(buf, chars);
}

void NmdcHub::revConnectToMe(const OnlineUser& aUser) {
	checkstate(); char buf[256];
	dcdebug("NmdcHub::revConnectToMe %s\n", aUser.getIdentity().getNick().c_str());
	int chars = sprintf_s(buf, 256, "$RevConnectToMe %s %s|", fromUtf8(getCurrentNick()).c_str(), fromUtf8(aUser.getIdentity().getNick()).c_str());
	Identity bUser = aUser.getIdentity();
	bUser.setFlag(Identity::RCTM_SND);
	if(chars != -1) {
		send(buf, chars);
	}
}

void NmdcHub::hubMessage(const string& aMessage) {
	checkstate();
	char buf[256];
	int iLen = sprintf_s(buf, 256, "<%s> ", fromUtf8(getCurrentNick()).c_str());
	if(iLen != -1) {
	   send(string(buf, iLen)+fromUtf8(escape(aMessage, true))+"|");
    }
}

void NmdcHub::myInfo() {
	if(state != STATE_LOGGEDIN && state != STATE_MYINFO) {
		return;
	}

	reloadSettings(false);

	dcdebug("MyInfo %s...\n", getCurrentNick().c_str());
	char StatusMode = '\x01';
	string tmp1 = ";**\x1fU9";
	string tmp2 = "+L9";
	string tmp3 = "+G9";
	string tmp4 = "+R9";
	string tmp5 = "+K9";
	string::size_type i;

	for(i = 0; i < 6; i++) {
		tmp1[i]++;
	}
	for(i = 0; i < 3; i++) {
		tmp2[i]++; tmp3[i]++; tmp4[i]++; tmp5[i]++;
	}
	char modeChar = '?';
	switch(SETTING(CONNECTION_TYPE)) {
        case SettingsManager::CONNECTION_ACTIVE:
    		modeChar = 'A';
    		break;
    	case SettingsManager::CONNECTION_PASSIVE:
		    modeChar = 'P';
		    break;
    	case SettingsManager::CONNECTION_SOCKS5:
		    modeChar = '5';
		    break;
		default:
            // ?!?
            break;
    }

	char tag[256];
	string connection = (BOOLSETTING(USE_UPLOAD_SPEED) || getStealth()) ? SETTING(UPLOAD_SPEED) : SETTING(CONNECTION);
	if (SETTING(THROTTLE_ENABLE) && SETTING(MAX_UPLOAD_SPEED_LIMIT) != 0) {
		sprintf_s(tag, 256, "%s" DCPPVERSIONSTRING "%s%c%s%s%s%s%s%s>", tmp1.c_str(), tmp2.c_str(), modeChar, tmp3.c_str(), getCounts().c_str(),
            tmp4.c_str(), Util::toString(Util::getSlots()).c_str(), tmp5.c_str(), Util::toString(SETTING(MAX_UPLOAD_SPEED_LIMIT)).c_str());
	} else {
		sprintf_s(tag, 256, "%s" DCPPVERSIONSTRING "%s%c%s%s%s%s>", tmp1.c_str(), tmp2.c_str(), modeChar, tmp3.c_str(), getCounts().c_str(),
            tmp4.c_str(), Util::toString(Util::getSlots()).c_str());
	}
	if(getStealth() == false) {
		if(UploadManager::getFireballStatus()) {
            StatusMode |= 0x8;
		} else if(UploadManager::getFileServerStatus()) {
			StatusMode |= 0x4;
		}

		if(Util::getAway()) {
			StatusMode |= 0x2;
		}
	}

	char myinfo[512];
	sprintf_s(myinfo, 512, "$MyINFO $ALL %s %s%s$ $%s%c$%s$", fromUtf8(getCurrentNick()).c_str(), 
		fromUtf8(escape(getCurrentDescription())).c_str(), tag, connection.c_str(), StatusMode, 
		fromUtf8(escape(SETTING(EMAIL))).c_str());

	int64_t newbytesshared = ShareManager::getInstance()->getShareSize();
	if(strcmp(myinfo, lastmyinfo.c_str()) != 0 || newbytesshared < (lastbytesshared - 1048576) || newbytesshared > (lastbytesshared + 1048576)) {
		lastmyinfo = myinfo;
		lastbytesshared = newbytesshared;
		tag[0] = NULL;
		sprintf_s(tag, 256, "%I64d$|", newbytesshared);
		strcat_s(myinfo, 512, tag);
		send(myinfo);
	}
}

void NmdcHub::GetUserList() {
	OnlineUser::List v;
	
	Lock l(cs);
	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
		v.push_back(i->second);
	}
	fire(ClientListener::UsersUpdated(), this, v);
}

void NmdcHub::search(int aSizeType, int64_t aSize, int aFileType, const string& aString) {
	checkstate(); 
	char c1 = (aSizeType == SearchManager::SIZE_DONTCARE) ? 'F' : 'T';
	char c2 = (aSizeType == SearchManager::SIZE_ATLEAST) ? 'F' : 'T';
	string tmp = ((aFileType == SearchManager::TYPE_TTH) ? "TTH:" + aString : fromUtf8(escape(aString)));

	for(size_t i = 0; i < tmp.size(); i++) {
        if(tmp[i] == ' ') {
            tmp[i] = '$';
        }
	}

	if(ClientManager::getInstance()->isActive() && (!BOOLSETTING(SEARCH_PASSIVE))) {
		string x = ClientManager::getInstance()->getCachedIp();
		const size_t BUF_SIZE = x.length() + aString.length() + 64;
		AutoArray<char> buf(BUF_SIZE);
		int chars = sprintf_s(buf, BUF_SIZE, "$Search %s:%d %c?%c?%I64d?%d?%s|", x.c_str(), (int)SearchManager::getInstance()->getPort(), c1, c2, aSize, aFileType+1, tmp.c_str());
		send(buf, chars);
	} else {
		const size_t BUF_SIZE = getCurrentNick().length() + aString.length() + 64;
		AutoArray<char> buf(BUF_SIZE);
		int chars = sprintf_s(buf, BUF_SIZE, "$Search Hub:%s %c?%c?%I64d?%d?%s|", fromUtf8(getCurrentNick()).c_str(), c1, c2, aSize, aFileType+1, tmp.c_str());
		send(buf, chars);
	}
}

string NmdcHub::escape(string str, bool bChat/* = false*/) {
    string::size_type i = 0;

	while( (i = str.find("&amp;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}
	i = 0;
	while( (i = str.find("&#36;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}
	i = 0;
	while( (i = str.find("&#124;", i)) != string::npos) {
		str.replace(i, 1, "&amp;");
		i += 4;
	}

	if(bChat == false) {
    	i = 0;
    	while( (i = str.find('$', i)) != string::npos) {
    		str.replace(i, 1, "&#36;");
    		i += 4;
    	}
    }

	i = 0;
	while( (i = str.find('|', i)) != string::npos) {
		str.replace(i, 1, "&#124;");
		i += 5;
	}

	return str;
}

void NmdcHub::privateMessage(const OnlineUser& aUser, const string& aMessage) {
	checkstate();

	send("$To: " + fromUtf8(aUser.getIdentity().getNick()) + " From: " + fromUtf8(getCurrentNick()) + " $" +
//        "<" + fromUtf8(getCurrentNick()) + "> " + escape(aMessage, true) + "|");
        fromUtf8(escape("<" + getCurrentNick() + "> " + aMessage)) + "|");
}

void NmdcHub::clearFlooders(uint64_t aTick) {
	while(!seekers.empty() && seekers.front().second + (5 * 1000) < aTick) {
		seekers.pop_front();
	}

	while(!flooders.empty() && flooders.front().second + (120 * 1000) < aTick) {
		flooders.pop_front();
	}
}

void NmdcHub::on(Connected) throw() {
	Client::on(Connected());

	unsetFlag(SUPPORTS_USERCOMMAND);
	unsetFlag(SUPPORTS_NOGETINFO);
	unsetFlag(SUPPORTS_USERIP2);
	unsetFlag(SUPPORTS_QUICKLIST);

	lastmyinfo.clear();
	lastbytesshared = 0;
	validatenicksent = false;
	bFirstOpList = true;
	PtokaX = false;
    ShittYnHub = false;
}

void NmdcHub::on(Failed, const string& aLine) throw() {
	clearUsers();
	Client::on(Failed(), aLine);
	updateCounts(true);
}

void NmdcHub::on(Second, uint64_t aTick) throw() {
	Client::on(Second(), aTick);

	if(state == STATE_LOGGEDIN && (getLastActivity() + getReconnDelay() * 1000) < aTick) {
		send("|", 1);
	} else if(state == STATE_DISCONNECTED && getAutoReconnect() && (aTick > (getLastActivity() + getReconnDelay() * 1000))) {
		// Try to reconnect...
		connect();
	} else if(state == STATE_LOCK && (getLastActivity() + 30000) < aTick) {
        socket->disconnect(false);
    }
}

OnlineUser* NmdcHub::findUserInChat(char * sText, bool &bNormal) {
    string nick = Util::emptyString;
	char *sTmp;
   
    switch(sText[0]) {
        case '<':
            if((sTmp = strchr(sText+1, '>')) != NULL) {
                nick = string(sText+1, sTmp-sText-1);
                if(sTmp[1] == ' ' && sTmp[2] == '/' && tolower(sTmp[3]) == 'm' && tolower(sTmp[4]) == 'e' && sTmp[5] == ' ' && sTmp[6] != NULL) {
                    for(size_t iNickLen = (sTmp - sText); iNickLen > 1; iNickLen--) {
                        sText[iNickLen+4] = sText[iNickLen-1];
                    }
    				sText[5] = ' ';
    				sText[4] = '*';
    				bNormal = false;
                }
            }
            break;
        case '*':
            if(sText[1] == ' ' && (sTmp = strchr(sText+2, ' ')) != NULL) {
                nick = string(sText+2, sTmp-sText-2);
            }
            break;
        default:
            break;
    }

    if(nick.size() != 0) {
        return findUser(TO_UTF8(Util::unescape(nick)));
    } else {
        return NULL;
    }
}

string NmdcHub::scanNickPrefix(const string& prefix) {
	string maxPrefix;
	tabCompleteNicks.clear();
	{
    	Lock l(cs);
    	for(NickMap::const_iterator i = users.begin(); i != users.end(); ++i) {
    		string prevNick, nick = i->second->getIdentity().getNick(), wholeNick = nick;

    		do {
    			string::size_type lp = prefix.size(), ln = nick.size();
    			if((ln >= lp) && (!Util::strnicmp(nick, prefix, lp))) {
    				if(maxPrefix == Util::emptyString)
                        maxPrefix = nick;	//ugly hack

    				tabCompleteNicks.push_back(nick);
    				tabCompleteNicks.push_back(wholeNick);
    				maxPrefix = maxPrefix.substr(0, 
                        mismatch(maxPrefix.begin(), maxPrefix.begin()+min(maxPrefix.size(), nick.size()), nick.begin(), compareCharsNoCase).first - maxPrefix.begin());
    			}

    			prevNick = nick;
    			nick = stripNick(nick);
    		} while (prevNick != nick);
    	}
    }

	return maxPrefix;
}

string NmdcHub::stripNick(const string& nick) const {
	if(nick.substr(0, 1) != "[")
        return nick;

	string::size_type x = nick.find(']');
	string ret;
	// Avoid full deleting of [IMCOOL][CUSIHAVENOTHINGELSETHANBRACKETS]-type nicks
	if((x != string::npos) && (nick.substr(x+1).length() > 0)) {
		ret = nick.substr(x+1);
	} else {
		ret = nick;
	}
	return ret;
}

/*
 * Some parts Copyright (C) 2002-2011 Petr Kozelka, PPK at czdc dot org
 * Licensed under GNU General Public License version 3.
 */
