/*
 * Copyright (C) 2001-2011 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdinc.h"
#include "DCPlusPlus.h"

#include "Util.h"
#include "File.h"

#include "ResourceManager.h"
#include "StringTokenizer.h"
#include "SettingsManager.h"
#include "version.h"
#include "File.h"
#include "SimpleXML.h"

#ifndef _WIN32
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/utsname.h>
#include <ctype.h>
#endif
#include <locale.h>

int64_t Util::mUptimeSeconds = 0;
string Util::emptyString;
wstring Util::emptyStringW;
tstring Util::emptyStringT;

string Util::awayMsg;
time_t Util::awayTime;

Util::CountryList Util::countries;

string Util::configPath;

int Util::OsMajor = 0;
int Util::OsMinor = 0;
int Util::slots;

static void sgenrand(unsigned long seed);

extern "C" void bz_internal_error(int errcode) {
	dcdebug("bzip2 internal error: %d\n", errcode);
}

#if defined(_WIN32) && (_MSC_VER >= 1400)
void WINAPI invalidParameterHandler(const wchar_t*, const wchar_t*, const wchar_t*, unsigned int, uintptr_t) {
	//do nothing, this exist because vs2k5 crt needs it not to crash on errors.
}
#endif

void Util::initialize() {
	Text::initialize();

	sgenrand((unsigned long)time(NULL));

	OSVERSIONINFOEX ver;
	memset(&ver, 0, sizeof(OSVERSIONINFOEX));
	ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

	if(!GetVersionEx((OSVERSIONINFO*)&ver)) {
		ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFO);
	}

	GetVersionEx((OSVERSIONINFO*)&ver);
	ver.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);
	OsMajor = ver.dwMajorVersion;
	OsMinor = ver.dwMinorVersion;

	TCHAR buf[MAX_PATH+1];
	::GetModuleFileName(NULL, buf, MAX_PATH);

	// System config path is CzDC executable path...
	string systemPath = Util::getFilePath(Text::fromT(buf));

	configPath = systemPath;

    if(File::FileExist(systemPath + "CzDC.xml") == false) { // We don't have config in app dir, use cur_user\appdata dir for config
        GetSystemDirectory(0, configPath);

        if(configPath != systemPath && File::FileExist(configPath + "CzDC.xml") == false && File::FileExist(configPath + "ip-to-country.csv") == false &&
            File::FileExist(systemPath + "ip-to-country.csv") == true) {
            File::ensureDirectory(configPath);
            ::MoveFile(Text::toT(systemPath + "ip-to-country.csv").c_str(), Text::toT(configPath + "ip-to-country.csv").c_str());
        }
    }

#if _MSC_VER >= 1400
	_set_invalid_parameter_handler(reinterpret_cast<_invalid_parameter_handler>(invalidParameterHandler));
#endif

    /*
     * This application uses the IP-to-Country Database
     * provided by WebHosting.Info (http://www.webhosting.info),
     * available from http://ip-to-country.webhosting.info.
    */

	string file = configPath + "ip-to-country.csv";
	FILE * ip2country = fopen(file.c_str(), "r");

	if(ip2country == NULL) {
		return;
	}

	char sLine[1024];

	CountryIter last = countries.end();
	uint32_t ui32StartIP = 0, ui32EndIP = 0, ui32EndIPprev = 0;
	uint16_t ui16Country = 0;

	while(fgets(sLine, 1024, ip2country) != NULL) {
		if(sLine[0] != '\"') {
			continue;
		}

		char * sStart = sLine+1;
		uint8_t ui8d = 0;

		size_t iLineLen = strlen(sLine);

		for(uint16_t ui16i = 1; ui16i < iLineLen; ui16i++) {
			if(sLine[ui16i] == '\"') {
				sLine[ui16i] = '\0';
				if(ui8d == 0) {
					ui32StartIP = toUInt32(sStart);
				} else if(ui8d == 1) {
					ui32EndIP = toUInt32(sStart);
				} else {
					ui16Country = ((uint16_t *)sStart)[0];
					break;
				}

				ui8d++;
				ui16i += (uint16_t)2;
				sStart = sLine+ui16i+1;
			}
		}

		if((ui32StartIP-1) != ui32EndIPprev) {
			last = countries.insert(last, make_pair((ui32StartIP-1), (uint16_t)16191));
		}
		last = countries.insert(last, make_pair(ui32EndIP, ui16Country));

		ui32EndIPprev = ui32EndIP;
	}

	fclose(ip2country);
}

string Util::validateMessage(string tmp, bool reverse, bool checkNewLines) {
	string::size_type i = 0;

	if(reverse) {
        // DC++ escapes
		while( (i = tmp.find("&#36;", i)) != string::npos) {
			tmp.replace(i, 5, "$");
			i++;
		}
		i = 0;
		while( (i = tmp.find("&#124;", i)) != string::npos) {
			tmp.replace(i, 6, "|");
			i++;
		}
		i = 0;
		while( (i = tmp.find("&amp;", i)) != string::npos) {
			tmp.replace(i, 5, "&");
			i++;
		}

        // Direct Connect escapes
    	i = 0;
    	while( (i = tmp.find("/%DCN036%/", i)) != string::npos) {
    		tmp.replace(i, 10, "$");
    		i++;
    	}
        i = 0;
    	while( (i = tmp.find("/%DCN096%/", i)) != string::npos) {
    		tmp.replace(i, 10, "`");
    		i++;
    	}
    	i = 0;
    	while( (i = tmp.find("/%DCN124%/", i)) != string::npos) {
    		tmp.replace(i, 10, "|");
    		i++;
    	}
    	i = 0;
    	while( (i = tmp.find("/%DCN126%/", i)) != string::npos) {
    		tmp.replace(i, 10, "~");
    		i++;
    	}

		if(checkNewLines) {
			// Check all '<' and '[' after newlines...
			i = 0;
			while( (i = tmp.find('\n', i)) != string::npos) {
				if(i + 1 < tmp.length()) {
					if(tmp[i+1] == '[' || tmp[i+1] == '<') {
						tmp.insert(i+1, "- ");
						i += 2;
					}
				}
				i++;
			}
		}
	} else {
		while( (i = tmp.find("&amp;", i)) != string::npos) {
			tmp.replace(i, 1, "&amp;");
			i += 4;
		}
		i = 0;
		while( (i = tmp.find("&#36;", i)) != string::npos) {
			tmp.replace(i, 1, "&amp;");
			i += 4;
		}
		i = 0;
		while( (i = tmp.find("&#124;", i)) != string::npos) {
			tmp.replace(i, 1, "&amp;");
			i += 4;
		}
		i = 0;
		while( (i = tmp.find('$', i)) != string::npos) {
			tmp.replace(i, 1, "&#36;");
			i += 4;
		}
		i = 0;
		while( (i = tmp.find('|', i)) != string::npos) {
			tmp.replace(i, 1, "&#124;");
			i += 5;
		}
	}
	return tmp;
}

#ifdef _WIN32
static const char badChars[] = {
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
	17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
	31, '<', '>', '/', '"', '|', '?', '*', 0
};
#else

static const char badChars[] = {
	1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16,
	17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
	31, '<', '>', '\\', '"', '|', '?', '*', 0
};
#endif

/**
 * Replaces all strange characters in a file with '_'
 * @todo Check for invalid names such as nul and aux...
 */
string Util::validateFileName(string tmp) {
	string::size_type i = 0;

	// First, eliminate forbidden chars
	while( (i = tmp.find_first_of(badChars, i)) != string::npos) {
		tmp[i] = '_';
		i++;
	}

	// Then, eliminate all ':' that are not the second letter ("c:\...")
	i = 0;
	while( (i = tmp.find(':', i)) != string::npos) {
		if(i == 1) {
			i++;
			continue;
		}
		tmp[i] = '_';
		i++;
	}

	// Remove the .\ that doesn't serve any purpose
	i = 0;
	while( (i = tmp.find("\\.\\", i)) != string::npos) {
		tmp.erase(i+1, 2);
	}
	i = 0;
	while( (i = tmp.find("/./", i)) != string::npos) {
		tmp.erase(i+1, 2);
	}

	// Remove any double \\ that are not at the beginning of the path...
	i = 1;
	while( (i = tmp.find("\\\\", i)) != string::npos) {
		tmp.erase(i+1, 1);
	}
	i = 1;
	while( (i = tmp.find("//", i)) != string::npos) {
		tmp.erase(i+1, 1);
	}

	// And last, but not least, the infamous ..\! ...
	i = 0;
	while( ((i = tmp.find("\\..\\", i)) != string::npos) ) {
		tmp[i + 1] = '_';
		tmp[i + 2] = '_';
		tmp[i + 3] = '_';
		i += 2;
	}
	i = 0;
	while( ((i = tmp.find("/../", i)) != string::npos) ) {
		tmp[i + 1] = '_';
		tmp[i + 2] = '_';
		tmp[i + 3] = '_';
		i += 2;
	}

	// Dots at the end of path names aren't popular
	i = 0;
	while( ((i = tmp.find(".\\", i)) != string::npos) ) {
		tmp[i] = '_';
		i += 1;
	}
	i = 0;
	while( ((i = tmp.find("./", i)) != string::npos) ) {
		tmp[i] = '_';
		i += 1;
	}


	return tmp;
}

string Util::cleanPathChars(string aNick) {
	string::size_type i = 0;

	while( (i = aNick.find_first_of("/.\\", i)) != string::npos) {
		aNick[i] = '_';
	}
	return aNick;
}

string Util::getShortTimeString(time_t t) {
	char buf[256];
	tm _tm;
	if(localtime_s(&_tm, &t)) {
		strcpy_s(buf, 256, "xx:xx");
	} else {
		strftime(buf, 254, SETTING(TIME_STAMPS_FORMAT).c_str(), &_tm);
	}
	return Text::toUtf8(buf);
}

/**
 * Decodes a URL the best it can...
 * Default ports:
 * http:// -> port 80
 * dchub:// -> port 411
 */
void Util::decodeUrl(const string& url, string& aServer, uint16_t& aPort, string& aFile) {
	// First, check for a protocol: xxxx://
	string::size_type i = 0, j, k;

	aServer = emptyString;
	aFile = emptyString;

	if( (j=url.find("://", i)) != string::npos) {
		// Protocol found
		string protocol = url.substr(0, j);
		i = j + 3;

		if(protocol == "http") {
			aPort = 80;
		} else if(protocol == "dchub") {
			aPort = 411;
		}
	}

	if( (j=url.find('/', i)) != string::npos) {
		// We have a filename...
		aFile = url.substr(j);
	}

	if( (k=url.find(':', i)) != string::npos) {
		// Port
		if(j == string::npos) {
			aPort = (uint16_t)Util::toInt(url.substr(k+1));
		} else if(k < j) {
			aPort = (uint16_t)Util::toInt(url.substr(k+1, j-k-1));
		}
	} else {
		k = j;
	}

	if(k == string::npos) {
		aServer = url.substr(i);
		if(i == 0) aPort = 411;
	} else
		aServer = url.substr(i, k-i);
}

string Util::getAwayMessage() {
	return (formatTime(awayMsg.empty() ? SETTING(DEFAULT_AWAY_MESSAGE) : awayMsg, awayTime));
}

string Util::formatBytes(int64_t aBytes) {
    static const char *units[] = {CSTRING(B), CSTRING(KB), CSTRING(MB), CSTRING(GB), CSTRING(TB), CSTRING(PB), CSTRING(EB), " ", " ", " ", " ", " ", " ", " ", " ", " "};
	char buf[128];
	if(aBytes < 1024 && aBytes > -1024) {
        sprintf_s(buf, 128, "%d %s", (int)(aBytes&0xffffffff), CSTRING(B));
	} else {
		long double ldBytes; int iter;
		for(ldBytes = (long double)aBytes, iter = 0; ldBytes > 1024 || ldBytes < -1024; iter++)
			ldBytes /= 1024;
		sprintf_s(buf, 128, "%0.2Lf %s", ldBytes, units[iter]);
	}
	return buf;
}

string Util::formatExactSize(int64_t aBytes) {
#ifdef _WIN32
		TCHAR buf[64];
		TCHAR number[64];
		NUMBERFMT nf;
		_stprintf_s(number, 64, _T("%I64d"), aBytes);
		TCHAR Dummy[16];

		/*No need to read these values from the system because they are not
		used to format the exact size*/
		nf.NumDigits = 0;
		nf.LeadingZero = 0;
		nf.NegativeOrder = 0;
		nf.lpDecimalSep = _T(",");

		GetLocaleInfo( LOCALE_SYSTEM_DEFAULT, LOCALE_SGROUPING, Dummy, 16 );
		nf.Grouping = _tstoi(Dummy);
		GetLocaleInfo( LOCALE_SYSTEM_DEFAULT, LOCALE_STHOUSAND, Dummy, 16 );
		nf.lpThousandSep = Dummy;

		GetNumberFormat(LOCALE_USER_DEFAULT, 0, number, &nf, buf, sizeof(buf)/sizeof(buf[0]));

		_stprintf_s(buf, 64, _T("%s %s"), buf, CTSTRING(B));
		return Text::fromT(buf);
#else
		char buf[64];
		snprintf(buf, 64, "%'lld", (long long int)aBytes);
		return string(buf) + STRING(B);
#endif
}

string Util::getLocalIp() {
	char buf[256];
	gethostname(buf, 256);

	struct addrinfo hints = { 0 };
	hints.ai_family = AF_INET;

    struct addrinfo *res;

	if(getaddrinfo(buf, NULL, &hints, &res) != 0 || res->ai_family != AF_INET) {
        return Util::emptyString;
    }

	string tmp = inet_ntoa(((sockaddr_in *)(res->ai_addr))->sin_addr);

    // We take the first ip as default, but if we can find a better one, use it instead...
	if(Util::isPrivateIp(tmp) || strncmp(tmp.c_str(), "169", 3) == 0) {
        struct addrinfo *nextAddrInfo = res->ai_next;

		while(nextAddrInfo != NULL) {
            struct addrinfo *curAddrInfo = nextAddrInfo;
            nextAddrInfo = curAddrInfo->ai_next;

            if(curAddrInfo->ai_family != AF_INET) {
                continue;
            }

			string tmp2 = inet_ntoa(((sockaddr_in *)(curAddrInfo->ai_addr))->sin_addr);

			if(!Util::isPrivateIp(tmp2) && strncmp(tmp2.c_str(), "169", 3) != 0) {
				tmp = tmp2;
			}
		}
	}

	freeaddrinfo(res);

	return tmp;
}

bool Util::isPrivateIp(string const& ip) {
	struct in_addr addr;

	addr.s_addr = inet_addr(ip.c_str());

	if(addr.s_addr != INADDR_NONE) {
		unsigned long haddr = ntohl(addr.s_addr);
		return ((haddr & 0xff000000) == 0x0a000000 || // 10.0.0.0/8
				(haddr & 0xff000000) == 0x7f000000 || // 127.0.0.0/8
				(haddr & 0xfff00000) == 0xac100000 || // 172.16.0.0/12
				(haddr & 0xffff0000) == 0xc0a80000);  // 192.168.0.0/16
	}
	return false;
}

typedef const uint8_t* ccp;
static wchar_t utf8ToLC(ccp& str) {
	wchar_t c = 0;
	if(str[0] & 0x80) {
		if(str[0] & 0x40) {
			if(str[0] & 0x20) {
				if(str[1] == 0 || str[2] == 0 ||
					!((((unsigned char)str[1]) & ~0x3f) == 0x80) ||
					!((((unsigned char)str[2]) & ~0x3f) == 0x80))
				{
					str++;
					return 0;
				}
				c = ((wchar_t)(unsigned char)str[0] & 0xf) << 12 |
					((wchar_t)(unsigned char)str[1] & 0x3f) << 6 |
					((wchar_t)(unsigned char)str[2] & 0x3f);
				str += 3;
			} else {
				if(str[1] == 0 ||
					!((((unsigned char)str[1]) & ~0x3f) == 0x80))
				{
					str++;
					return 0;
				}
				c = ((wchar_t)(unsigned char)str[0] & 0x1f) << 6 |
					((wchar_t)(unsigned char)str[1] & 0x3f);
				str += 2;
			}
		} else {
			str++;
			return 0;
		}
	} else {
		wchar_t c = Text::asciiToLower((char)str[0]);
		str++;
		return c;
	}

	return Text::toLower(c);
}

string::size_type Util::findSubString(const string& aString, const string& aSubString, string::size_type start) throw() {
	if(aString.length() < start)
		return (string::size_type)string::npos;

	if(aString.length() - start < aSubString.length())
		return (string::size_type)string::npos;

	if(aSubString.empty())
		return 0;

	// Hm, should start measure in characters or in bytes? bytes for now...
	const uint8_t* tx = (const uint8_t*)aString.c_str() + start;
	const uint8_t* px = (const uint8_t*)aSubString.c_str();

	const uint8_t* end = tx + aString.length() - start - aSubString.length() + 1;

	wchar_t wp = utf8ToLC(px);

	while(tx < end) {
		const uint8_t* otx = tx;
		if(wp == utf8ToLC(tx)) {
			const uint8_t* px2 = px;
			const uint8_t* tx2 = tx;

			for(;;) {
				if(*px2 == 0)
					return otx - (uint8_t*)aString.c_str();

				if(utf8ToLC(px2) != utf8ToLC(tx2))
					break;
			}
		}
	}

	return (string::size_type)string::npos;
}

wstring::size_type Util::findSubString(const wstring& aString, const wstring& aSubString, wstring::size_type pos) throw() {
	if(aString.length() < pos)
		return static_cast<wstring::size_type>(wstring::npos);

	if(aString.length() - pos < aSubString.length())
		return static_cast<wstring::size_type>(wstring::npos);

	if(aSubString.empty())
		return 0;

	wstring::size_type j = 0;
	wstring::size_type end = aString.length() - aSubString.length() + 1;

	for(; pos < end; ++pos) {
		if(Text::toLower(aString[pos]) == Text::toLower(aSubString[j])) {
			wstring::size_type tmp = pos+1;
			bool found = true;
			for(++j; j < aSubString.length(); ++j, ++tmp) {
				if(Text::toLower(aString[tmp]) != Text::toLower(aSubString[j])) {
					j = 0;
					found = false;
					break;
				}
			}

			if(found)
				return pos;
		}
	}
	return static_cast<wstring::size_type>(wstring::npos);
}

int Util::stricmp(const char* a, const char* b) {
	while(*a) {
		wchar_t ca = 0, cb = 0;
		int na = Text::utf8ToWc(a, ca);
		int nb = Text::utf8ToWc(b, cb);
		ca = Text::toLower(ca);
		cb = Text::toLower(cb);
		if(ca != cb) {
			return (int)ca - (int)cb;
		}
		a += abs(na);
		b += abs(nb);
	}
	wchar_t ca = 0, cb = 0;
	Text::utf8ToWc(a, ca);
	Text::utf8ToWc(b, cb);

	return (int)Text::toLower(ca) - (int)Text::toLower(cb);
}

int Util::strnicmp(const char* a, const char* b, size_t n) {
	const char* end = a + n;
	while(*a && a < end) {
		wchar_t ca = 0, cb = 0;
		int na = Text::utf8ToWc(a, ca);
		int nb = Text::utf8ToWc(b, cb);
		ca = Text::toLower(ca);
		cb = Text::toLower(cb);
		if(ca != cb) {
			return (int)ca - (int)cb;
		}
		a += abs(na);
		b += abs(nb);
	}
	wchar_t ca = 0, cb = 0;
	Text::utf8ToWc(a, ca);
	Text::utf8ToWc(b, cb);
	return (a >= end) ? 0 : ((int)Text::toLower(ca) - (int)Text::toLower(cb));
}

int Util::userstricmp(const wchar_t* a, const wchar_t* b) {
    wchar_t f, l;
	double v1, v2;
    do {
		v1 = 0; v2 = 0; 
		bool t1 = isNumeric(*a);
		bool t2 = isNumeric(*b);
		if(t1 != t2) return (t1) ? -1 : 1;
		if(t1) {
			while(isNumeric(*a)) {
            	v1 *= 10;
				v1 += *a - '0';
				a++;
			}
			
			if(*a == '.') {
                double iMod = 10;
                a++;
    			while(isNumeric(*a)) {
    				v1 += (*a - '0')/iMod;
    				iMod *= 10;
    				a++;
    			}
            }

			while(isNumeric(*b)) {
				v2 *= 10;
				v2 += *b - '0';
				b++;
			}
			
			if(*b == '.') {
                double iMod = 10;
                b++;
    			while(isNumeric(*b)) {
    				v2 += (*b - '0')/iMod;
    				iMod *= 10;
    				b++;
    			}
            }

			if(v1 != v2)
                return (v1 < v2) ? -1 : 1;
		}
		f = towlower((unsigned short)(*(a++)));
		l = towlower((unsigned short)(*(b++)));
    } while((f) && (f == l));
    return (int)(f - l);
}

string Util::encodeURI(const string& aString, bool reverse) {
	// reference: rfc2396
	string tmp = aString;
	if(reverse) {
		string::size_type idx;
		for(idx = 0; idx < tmp.length(); ++idx) {
			if(tmp.length() > idx + 2 && tmp[idx] == '%' && isxdigit(tmp[idx+1]) && isxdigit(tmp[idx+2])) {
				tmp[idx] = fromHexEscape(tmp.substr(idx+1,2));
				tmp.erase(idx+1, 2);
			} else { // reference: rfc1630, magnet-uri draft
				if(tmp[idx] == '+')
					tmp[idx] = ' ';
			}
		}
	} else {
		const string disallowed = ";/?:@&=+$," // reserved
								  "<>#%\" "    // delimiters
								  "{}|\\^[]`"; // unwise
		string::size_type idx, loc;
		for(idx = 0; idx < tmp.length(); ++idx) {
			if(tmp[idx] == ' ') {
				tmp[idx] = '+';
			} else {
				if(tmp[idx] <= 0x1F || tmp[idx] >= 0x7f || (loc = disallowed.find_first_of(tmp[idx])) != string::npos) {
					tmp.replace(idx, 1, toHexEscape(tmp[idx]));
					idx+=2;
				}
			}
		}
	}
	return tmp;
}

/**
 * This function takes a string and a set of parameters and transforms them according to
 * a simple formatting rule, similar to strftime. In the message, every parameter should be
 * represented by %[name]. It will then be replaced by the corresponding item in
 * the params stringmap. After that, the string is passed through strftime with the current
 * date/time and then finally written to the log file. If the parameter is not present at all,
 * it is removed from the string completely...
 */
string Util::formatParams(const string& msg, StringMap& params, bool filter) {
	string result = msg;

	string::size_type i, j, k;
	i = 0;
	while (( j = result.find("%[", i)) != string::npos) {
		if( (result.size() < j + 2) || ((k = result.find(']', j + 2)) == string::npos) ) {
			break;
		}
		string name = result.substr(j + 2, k - j - 2);
		StringMapIter smi = params.find(name);
		if(smi == params.end()) {
			result.erase(j, k-j + 1);
			i = j;
		} else {
			if(smi->second.find_first_of("%\\./") != string::npos) {
				string tmp = smi->second;	// replace all % in params with %% for strftime
				string::size_type m = 0;
				while(( m = tmp.find('%', m)) != string::npos) {
					tmp.replace(m, 1, "%%");
					m+=2;
				}
				if(filter) {
					// Filter chars that produce bad effects on file systems
					m = 0;
					while(( m = tmp.find_first_of("\\./", m)) != string::npos) {
						tmp[m] = '_';
					}
				}

				result.replace(j, k-j + 1, tmp);
				i = j + tmp.size();
			} else {
				result.replace(j, k-j + 1, smi->second);
				i = j + smi->second.size();
			}
		}
	}

	result = formatTime(result, time(NULL));

	return result;
}

/** Fix for wide formatting bug in wcsftime in the ms c lib for multibyte encodings of unicode in singlebyte locales */
string fixedftime(const string& format, struct tm* t) {
	string ret = format;
	static const char codes[] = "aAbBcdHIjmMpSUwWxXyYzZ%";
	static const size_t codeslen = 23;

	char tmp[4];
	tmp[0] = '%';
	tmp[1] = tmp[2] = tmp[3] = 0;

	StringMap sm;
	char buf[1024];
	for(size_t i = 0; i < codeslen; ++i) {
		tmp[1] = codes[i];
		tmp[2] = 0;
		strftime(buf, 1024-1, tmp, t);
		sm[tmp] = buf;

		tmp[1] = '#';
		tmp[2] = codes[i];
		strftime(buf, 1024-1, tmp, t);
		sm[tmp] = buf; 
	}

	for(StringMapIter i = sm.begin(); i != sm.end(); ++i) {
		for(string::size_type j = ret.find(i->first); j != string::npos; j = ret.find(i->first, j)) {
			ret.replace(j, i->first.length(), i->second);
			j += i->second.length() - i->first.length();
		}
	}

	return ret;
}

string Util::formatTime(const string &msg, const time_t t) {
	if (!msg.empty()) {
		size_t bufsize = msg.size() + 256;
		struct tm loc;

		if(localtime_s(&loc, &t)) {
			return Util::emptyString;
		}
#if _WIN32
		AutoArray<TCHAR> buf(bufsize);

		if(!_tcsftime(buf, bufsize-1, Text::toT(msg).c_str(), &loc)) {
			return fixedftime(msg, &loc);
		}

		return Text::fromT(tstring(buf));
#else
		// will this give wide representations for %a and %A?
		// surely win32 can't have a leg up on linux/unixen in this area. - Todd
		AutoArray<char> buf(bufsize);

		while(!strftime(buf, bufsize-1, msg.c_str(), loc)) {
			bufsize+=64;
			buf = new char[bufsize];
		}

		return Text::toUtf8(string(buf));
#endif
	}
	return Util::emptyString;
}

/* Below is a high-speed random number generator with much
   better granularity than the CRT one in msvc...(no, I didn't
   write it...see copyright) */
/* Copyright (C) 1997 Makoto Matsumoto and Takuji Nishimura.
   Any feedback is very welcome. For any question, comments,
   see http://www.math.keio.ac.jp/matumoto/emt.html or email
   matumoto@math.keio.ac.jp */
/* Period parameters */
#define N 624
#define M 397
#define MATRIX_A 0x9908b0df   /* constant vector a */
#define UPPER_MASK 0x80000000 /* most significant w-r bits */
#define LOWER_MASK 0x7fffffff /* least significant r bits */

/* Tempering parameters */
#define TEMPERING_MASK_B 0x9d2c5680
#define TEMPERING_MASK_C 0xefc60000
#define TEMPERING_SHIFT_U(y) (y >> 11)
#define TEMPERING_SHIFT_S(y) (y << 7)
#define TEMPERING_SHIFT_T(y) (y << 15)
#define TEMPERING_SHIFT_L(y) (y >> 18)

static unsigned long mt[N]; /* the array for the state vector  */
static int mti=N+1; /* mti==N+1 means mt[N] is not initialized */

/* initializing the array with a NONZERO seed */
static void sgenrand(unsigned long seed) {
	/* setting initial seeds to mt[N] using         */
	/* the generator Line 25 of Table 1 in          */
	/* [KNUTH 1981, The Art of Computer Programming */
	/*    Vol. 2 (2nd Ed.), pp102]                  */
	mt[0]= seed & 0xffffffff;
	for (mti=1; mti<N; mti++)
		mt[mti] = (69069 * mt[mti-1]) & 0xffffffff;
}

uint32_t Util::rand() {
	unsigned long y;
	static unsigned long mag01[2]={0x0, MATRIX_A};
	/* mag01[x] = x * MATRIX_A  for x=0,1 */

	if (mti >= N) { /* generate N words at one time */
		int kk;

		if (mti == N+1)   /* if sgenrand() has not been called, */
			sgenrand(4357); /* a default initial seed is used   */

		for (kk=0;kk<N-M;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+M] ^ (y >> 1) ^ mag01[y & 0x1];
		}
		for (;kk<N-1;kk++) {
			y = (mt[kk]&UPPER_MASK)|(mt[kk+1]&LOWER_MASK);
			mt[kk] = mt[kk+(M-N)] ^ (y >> 1) ^ mag01[y & 0x1];
		}
		y = (mt[N-1]&UPPER_MASK)|(mt[0]&LOWER_MASK);
		mt[N-1] = mt[M-1] ^ (y >> 1) ^ mag01[y & 0x1];

		mti = 0;
	}

	y = mt[mti++];
	y ^= TEMPERING_SHIFT_U(y);
	y ^= TEMPERING_SHIFT_S(y) & TEMPERING_MASK_B;
	y ^= TEMPERING_SHIFT_T(y) & TEMPERING_MASK_C;
	y ^= TEMPERING_SHIFT_L(y);

	return y;
}

/*
 *	getIpCountry
 *	This function returns the country(Abbreviation) of an ip
 *	for exemple: it returns "PT", whitch standards for "Portugal"
*/
string Util::getIpCountry (string IP) {
	if (BOOLSETTING(GET_USER_COUNTRY)) {
		dcassert(count(IP.begin(), IP.end(), '.') == 3);

		//e.g IP 23.24.25.26 : w=23, x=24, y=25, z=26
		string::size_type a = IP.find('.');
		string::size_type b = IP.find('.', a+1);
		string::size_type c = IP.find('.', b+2);

		uint32_t ipnum = (Util::toUInt32(IP.c_str()) << 24) |
			(Util::toUInt32(IP.c_str() + a + 1) << 16) |
			(Util::toUInt32(IP.c_str() + b + 1) << 8) |
			(Util::toUInt32(IP.c_str() + c + 1) );

		CountryIter i = countries.lower_bound(ipnum);

		if(i != countries.end()) {
			return string((char*)&(i->second), 2);
		}
	}

	return Util::emptyString; //if doesn't returned anything already, something is wrong...
}

string Util::toDOS(const string& tmp) {
	if(tmp.empty())
		return Util::emptyString;

	string tmp2(tmp);

	if(tmp2[0] == '\r' && (tmp2.size() == 1 || tmp2[1] != '\n')) {
		tmp2.insert(1, "\n");
	}
	for(string::size_type i = 1; i < tmp2.size() - 1; ++i) {
		if(tmp2[i] == '\r' && tmp2[i+1] != '\n') {
			// Mac ending
			tmp2.insert(i+1, "\n");
			i++;
		} else if(tmp2[i] == '\n' && tmp2[i-1] != '\r') {
			// Unix encoding
			tmp2.insert(i, "\r");
			i++;
		}
	}
	return tmp2;
}

string Util::formatMessage(const string& nick, const string& message) {
	string tmp = '<' + nick + "> " + message;
	// Check all '<' and '[' after newlines as they're probably pasts...
	size_t i = 0;
	while( (i = tmp.find('\n', i)) != string::npos) {
		if(i + 1 < tmp.length()) {
			if(tmp[i+1] == '[' || tmp[i+1] == '<') {
				tmp.insert(i+1, "- ");
				i += 2;
			}
		}
		i++;
	}
	return toDOS(tmp);
}

string Util::getTimeString() {
	char buf[64];
	time_t _tt; tm _tm;
	time(&_tt);
	if(localtime_s(&_tm, &_tt)) {
		strcpy_s(buf, 64, "xx:xx:xx");
	} else {
		strftime(buf, 64, "%X", &_tm);
	}
	return buf;
}

string Util::toAdcFile(const string& file) {
	if(file == "files.xml.bz2" || file == "MyList.DcLst")
		return file;

	string ret;
	ret.reserve(file.length() + 1);
	ret += '/';
	ret += file;
	for(string::size_type i = 0; i < ret.length(); ++i) {
		if(ret[i] == '\\') {
			ret[i] = '/';
		}
	}
	return ret;
}
string Util::toNmdcFile(const string& file) {
	if(file.empty())
		return Util::emptyString;

	string ret(file.substr(1));
	for(string::size_type i = 0; i < ret.length(); ++i) {
		if(ret[i] == '/') {
			ret[i] = '\\';
		}
	}
	return ret;
}

bool Util::getAway() {
	return BOOLSETTING(AWAY_ENABLED);
}

void Util::setAway(bool aAway) {
	SettingsManager::getInstance()->set(SettingsManager::AWAY_ENABLED, aAway);
	if(aAway)
		awayTime = time(NULL);
}

int Util::getSlots() {
	return slots;
}

void Util::setSlots(int Slots) {
	slots = Slots;
}

char* Util::stristr(const char *str1, const char *str2) {
	char *s1, *s2;
	char *cp = (char *)str1;
	if(!*str2)
		return (char *)str1;
	while(*cp) {
		s1 = cp;
		s2 = (char *) str2;
		while(*s1 && *s2 && (!(*s1-*s2) ||
			!(*s1-tolower(*s2)) || !(*s1-toupper(*s2))))
				s1++, s2++;
		if(!*s2) {
			return cp;
		}
		cp++;
	}
	return NULL;
}

TCHAR* Util::strstr(const TCHAR *str1, const TCHAR *str2, int &pnIdxFound) {
	TCHAR *s1, *s2;
	TCHAR *cp = (TCHAR *)str1;
	if(!*str2)
		return (TCHAR *)str1;

	int nIdx = 0;
	while(*cp) {
		s1 = cp;
		s2 = (TCHAR *) str2;
                while(*s1 && *s2 && !(*s1-*s2))
                        s1++, s2++;
		if(!*s2) {
			pnIdxFound = nIdx;
			return cp;
		}
		cp++;
		nIdx++;
	}

	pnIdxFound = -1;
	return NULL;
}

static const string base64_chars = 
    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    "abcdefghijklmnopqrstuvwxyz"
    "0123456789+/";

static inline bool is_base64(unsigned char c) {
    return (isalnum(c) || (c == '+') || (c == '/'));
}

string Util::base64_encode(unsigned char const* bytes_to_encode, size_t in_len) {
    string ret;
    int i = 0;
    int j = 0;
    unsigned char char_array_3[3];
    unsigned char char_array_4[4];

    while(in_len--) {
        char_array_3[i++] = *(bytes_to_encode++);
        if(i == 3) {
            char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
            char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
            char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
            char_array_4[3] = char_array_3[2] & 0x3f;

            for(i = 0; (i <4) ; i++)
                ret += base64_chars[char_array_4[i]];

                i = 0;
        }
    }

    if(i) {
        for(j = i; j < 3; j++)
            char_array_3[j] = '\0';

        char_array_4[0] = (char_array_3[0] & 0xfc) >> 2;
        char_array_4[1] = ((char_array_3[0] & 0x03) << 4) + ((char_array_3[1] & 0xf0) >> 4);
        char_array_4[2] = ((char_array_3[1] & 0x0f) << 2) + ((char_array_3[2] & 0xc0) >> 6);
        char_array_4[3] = char_array_3[2] & 0x3f;

        for (j = 0; (j < i + 1); j++)
            ret += base64_chars[char_array_4[j]];

        while((i++ < 3))
            ret += '=';

    }

    return ret;
}

string Util::base64_decode(string const& encoded_string) {
    size_t in_len = encoded_string.size();
    int i = 0;
    int j = 0;
    int in_ = 0;
    unsigned char char_array_4[4], char_array_3[3];
    string ret;

    while(in_len-- && ( encoded_string[in_] != '=') && is_base64(encoded_string[in_])) {
        char_array_4[i++] = encoded_string[in_]; in_++;
        if(i ==4) {
            for (i = 0; i <4; i++)
                char_array_4[i] = (char)base64_chars.find(char_array_4[i]);

            char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
            char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
            char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

            for(i = 0; (i < 3); i++)
                ret += char_array_3[i];

            i = 0;
        }
    }

    if(i) {
        for(j = i; j <4; j++)
            char_array_4[j] = 0;

        for(j = 0; j <4; j++)
            char_array_4[j] = (char)base64_chars.find(char_array_4[j]);

        char_array_3[0] = (char_array_4[0] << 2) + ((char_array_4[1] & 0x30) >> 4);
        char_array_3[1] = ((char_array_4[1] & 0xf) << 4) + ((char_array_4[2] & 0x3c) >> 2);
        char_array_3[2] = ((char_array_4[2] & 0x3) << 6) + char_array_4[3];

        for (j = 0; (j < i - 1); j++)
            ret += char_array_3[j];
    }

    return ret;
}

TCHAR* Util::tsring2tcharp(TCHAR * srctchr, tstring const& srctstr) {
    if(srctchr != NULL) {
        if(_tcscmp(srctchr, srctstr.c_str()) == NULL) {
            return srctchr;
        }
		delete [] srctchr;
		srctchr = NULL;
	}

    if(srctstr.size() > 0) {
        srctchr = new TCHAR[srctstr.size()+1];
        if(srctchr != NULL) {
            _tcsncpy(srctchr, srctstr.c_str(), srctstr.size());
            srctchr[srctstr.size()] = _T('\0');
        }
    }
    return srctchr;
}

string Util::unescape(string str) {
    string::size_type i = 0;

    // DC++ escapes
    while( (i = str.find("&#36;", i)) != string::npos) {
        str.replace(i, 5, "$");
		i++;
	}
	i = 0;
	while( (i = str.find("&#124;", i)) != string::npos) {
		str.replace(i, 6, "|");
		i++;
	}
	i = 0;
	while( (i = str.find("&amp;", i)) != string::npos) {
		str.replace(i, 5, "&");
		i++;
	}

    // Direct Connect escapes
	i = 0;
	while( (i = str.find("/%DCN036%/", i)) != string::npos) {
		str.replace(i, 10, "$");
		i++;
	}
	i = 0;
	while( (i = str.find("/%DCN096%/", i)) != string::npos) {
		str.replace(i, 10, "`");
		i++;
	}
	i = 0;
	while( (i = str.find("/%DCN124%/", i)) != string::npos) {
		str.replace(i, 10, "|");
		i++;
	}
	i = 0;
	while( (i = str.find("/%DCN126%/", i)) != string::npos) {
		str.replace(i, 10, "~");
		i++;
	}

	return str;
}

string Util::getListPath() {
    return SETTING(DOWNLOAD_DIRECTORY) + "CzDC FileLists" PATH_SEPARATOR_STR;
}

void Util::GetSystemDirectory(uint8_t ui8ID, string &sPath) {
    if(OsMajor >= 6) { //Vista and higher
        HINSTANCE hShell32 = ::LoadLibrary(_T("shell32.dll"));

        typedef HRESULT (WINAPI * SHGKFP)(REFKNOWNFOLDERID, DWORD, HANDLE, PWSTR *);
        SHGKFP pSHGKFP = (SHGKFP)::GetProcAddress(hShell32, "SHGetKnownFolderPath");

        PWSTR Path = NULL;

        switch(ui8ID) {
            case 0: // Config path
                if(pSHGKFP(FOLDERID_RoamingAppData, 0, NULL, &Path) == S_OK) {
                    sPath = Text::fromT(Path) + (PATH_SEPARATOR_STR "CzDC" PATH_SEPARATOR_STR);
                }

                break;
            case 1: // Logs path
                if(pSHGKFP(FOLDERID_Documents, 0, NULL, &Path) == S_OK) {
                    sPath = Text::fromT(Path) + (PATH_SEPARATOR_STR "CzDC Logs" PATH_SEPARATOR_STR);
                }

                break;
            case 2: // Downloads path
                if(pSHGKFP(FOLDERID_Downloads, 0, NULL, &Path) == S_OK) {
                    sPath = Text::fromT(Path) + PATH_SEPARATOR_STR;
                }

                break;
        }

        ::CoTaskMemFree(Path);

        ::FreeLibrary(hShell32);
    } else { // Prehistoric windoze
        TCHAR tcPath[MAX_PATH];

        if(ui8ID == 0) {
            if(::SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, SHGFP_TYPE_CURRENT, tcPath) == S_OK) {
                sPath = Text::fromT(tcPath) + (PATH_SEPARATOR_STR "CzDC" PATH_SEPARATOR_STR);
            }
        } else {
            if(::SHGetFolderPath(NULL, CSIDL_PERSONAL, NULL, SHGFP_TYPE_CURRENT, tcPath) == S_OK) {
                if(ui8ID == 1) {
                    sPath = Text::fromT(tcPath) + (PATH_SEPARATOR_STR "CzDC Logs" PATH_SEPARATOR_STR);
                } else if(ui8ID == 2) {
                    sPath = Text::fromT(tcPath) + (PATH_SEPARATOR_STR "Downloads" PATH_SEPARATOR_STR);
                }
            }
        }
    }
}

/*
 * Some parts Copyright (C) 2002-2011 Petr Kozelka, PPK at czdc dot org
 * Licensed under GNU General Public License version 3.
 */
